@@ -732,18 +732,25 @@ ip_defrag_stub(struct ip *iph, struct ip **defrag)
   return IPF_ISF;
 }
 
-int
+void
 ip_frag_init(int n)
 {
   struct timeval tv;
-  
+
   gettimeofday(&tv, 0);
   time0 = tv.tv_sec;
-  fragtable = (struct hostfrags **) malloc(n * sizeof(struct hostfrags *));
+  fragtable = (struct hostfrags **) calloc(n, sizeof(struct hostfrags *));
   if (!fragtable)
     nids_params.no_mem("ip_frag_init");
-  memset(fragtable, 0, n * sizeof(struct hostfrags *));
   hash_size = n;
+}
 
-  return 0;
+void
+ip_frag_exit(void)
+{
+  if (fragtable) {
+    free(fragtable);
+    fragtable = NULL;
+  }
+  /* FIXME: do we need to free anything else? */
 }
