@@ -27,6 +27,9 @@
 #include "tcp.h"
 #include "util.h"
 #include "nids.h"
+#ifdef HAVE_LIBGTHREAD_2_0
+#include <glib.h>
+#endif
 
 #ifdef __linux__
 extern int set_all_promisc();
@@ -43,12 +46,32 @@ static struct proc_node *ip_procs;
 static struct proc_node *udp_procs;
 
 struct proc_node *tcp_procs;
-static unsigned int linkoffset = 0;
 static int linktype;
-static pcap_t *desc = 0;
+static pcap_t *desc = NULL;
+
+#ifdef HAVE_LIBGTHREAD_2_0
+
+/* async queue for multiprocessing - mcree */
+static GAsyncQueue *cap_queue;
+
+/* items in the queue */
+struct cap_queue_item {
+     void *data;
+     bpf_u_int32 caplen;
+};
+
+/* marks end of queue */
+static struct cap_queue_item EOF_item;
+
+/* error buffer for glib calls */
+static GError *gerror = NULL;
+
+#endif
 
 char nids_errbuf[PCAP_ERRBUF_SIZE];
 struct pcap_pkthdr * nids_last_pcap_header = NULL;
+u_char *nids_last_pcap_data = NULL;
+u_int nids_linkoffset = 0;
 
 char *nids_warnings[] = {
     "Murphy - you never should see this message !",
@@ -80,7 +103,11 @@ struct nids_prm nids_params = {
     NULL,			/* pcap_filter */
     1,				/* promisc */
     0,				/* one_loop_less */
-    1024			/* pcap_timeout */
+    1024,			/* pcap_timeout */
+    0,				/* multiproc */
+    20000,			/* queue_limit */
+    0,				/* tcp_workarounds */
+    NULL			/* pcap_desc */
 };
 
 static int nids_ip_filter(struct ip *x, int len)
@@ -161,6 +188,17 @@ static void nids_syslog(int type, int errnum, struct ip *iph, void *data)
     }
 }
 
+/* called either directly from pcap_hand() or from cap_queue_process_thread()
+ * depending on the value of nids_params.multiproc - mcree
+ */
+static void call_ip_frag_procs(void *data,bpf_u_int32 caplen)
+{
+    struct proc_node *i;
+    for (i = ip_frag_procs; i; i = i->next)
+	(i->item) (data, caplen);
+}
+
+
 /* wireless frame types, mostly from tcpdump (wam) */
 #define FC_TYPE(fc)             (((fc) >> 2) & 0x3)
 #define FC_WEP(fc)              ((fc) & 0x4000)
@@ -178,15 +216,27 @@ static void nids_syslog(int type, int errnum, struct ip *iph, void *data)
 #define LLC_OFFSET_TO_TYPE_FIELD 6
 #define ETHERTYPE_IP 0x0800
 
-static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
+void nids_pcap_handler(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 {
-    struct proc_node *i;
     u_char *data_aligned;
+#ifdef HAVE_LIBGTHREAD_2_0
+    struct cap_queue_item *qitem;
+#endif
 #ifdef DLT_IEEE802_11
     unsigned short fc;
     int linkoffset_tweaked_by_prism_code = 0;
 #endif
+
+    /*
+     * Check for savagely closed TCP connections. Might
+     * happen only when nids_params.tcp_workarounds is non-zero;
+     * otherwise nids_tcp_timeouts is always NULL.
+     */
+    if (NULL != nids_tcp_timeouts)
+      tcp_check_timeouts(&hdr->ts);
+
     nids_last_pcap_header = hdr;
+    nids_last_pcap_data = data;
     (void)par; /* warnings... */
     switch (linktype) {
     case DLT_EN10MB:
@@ -195,10 +245,10 @@ static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 	/* Only handle IP packets and 802.1Q VLAN tagged packets below. */
 	if (data[12] == 8 && data[13] == 0) {
 	    /* Regular ethernet */
-	    linkoffset = 14;
+	    nids_linkoffset = 14;
 	} else if (data[12] == 0x81 && data[13] == 0) {
 	    /* Skip 802.1Q VLAN and priority information */
-	    linkoffset = 18;
+	    nids_linkoffset = 18;
 	} else
 	    /* non-ip frame */
 	    return;
@@ -208,7 +258,7 @@ static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 #error DLT_PRISM_HEADER is defined, but DLT_IEEE802_11 is not ???
 #endif
     case DLT_PRISM_HEADER:
-	linkoffset = 144; //sizeof(prism2_hdr);
+	nids_linkoffset = 144; //sizeof(prism2_hdr);
 	linkoffset_tweaked_by_prism_code = 1;
         //now let DLT_IEEE802_11 do the rest
 #endif
@@ -218,8 +268,8 @@ static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 	 * works for tcpdump, so who am I to complain? (wam)
 	 */
 	if (!linkoffset_tweaked_by_prism_code)
-		linkoffset = 0;
-	fc = EXTRACT_LE_16BITS(data + linkoffset);
+		nids_linkoffset = 0;
+	fc = EXTRACT_LE_16BITS(data + nids_linkoffset);
 	if (FC_TYPE(fc) != T_DATA || FC_WEP(fc)) {
 	    return;
 	}
@@ -227,26 +277,26 @@ static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 	    /* a wireless distribution system packet will have another
 	     * MAC addr in the frame
 	     */
-	    linkoffset += 30;
+	    nids_linkoffset += 30;
 	} else {
-	    linkoffset += 24;
+	    nids_linkoffset += 24;
 	}
-	if (hdr->len < linkoffset + LLC_FRAME_SIZE)
+	if (hdr->len < nids_linkoffset + LLC_FRAME_SIZE)
 	    return;
 	if (ETHERTYPE_IP !=
-	    EXTRACT_16BITS(data + linkoffset + LLC_OFFSET_TO_TYPE_FIELD)) {
+	    EXTRACT_16BITS(data + nids_linkoffset + LLC_OFFSET_TO_TYPE_FIELD)) {
 	    /* EAP, LEAP, and other 802.11 enhancements can be 
 	     * encapsulated within a data packet too.  Look only at
 	     * encapsulated IP packets (Type field of the LLC frame).
 	     */
 	    return;
 	}
-	linkoffset += LLC_FRAME_SIZE;
+	nids_linkoffset += LLC_FRAME_SIZE;
 	break;
 #endif
     default:;
     }
-    if (hdr->caplen < linkoffset)
+    if (hdr->caplen < nids_linkoffset)
 	return;
 
 /*
@@ -255,15 +305,43 @@ static void pcap_hand(u_char * par, struct pcap_pkthdr *hdr, u_char * data)
 * handle->offset in pcap sources), so memcpy should not be called.
 */
 #ifdef LBL_ALIGN
-    if ((unsigned long) (data + linkoffset) & 0x3) {
-	data_aligned = alloca(hdr->caplen - linkoffset + 4);
+    if ((unsigned long) (data + nids_linkoffset) & 0x3) {
+	data_aligned = alloca(hdr->caplen - nids_linkoffset + 4);
 	data_aligned -= (unsigned long) data_aligned % 4;
-	memcpy(data_aligned, data + linkoffset, hdr->caplen - linkoffset);
-    } else
+	memcpy(data_aligned, data + nids_linkoffset, hdr->caplen - nids_linkoffset);
+    } else 
 #endif
-	data_aligned = data + linkoffset;
-    for (i = ip_frag_procs; i; i = i->next)
-	(i->item) (data_aligned, hdr->caplen - linkoffset);
+  data_aligned = data + nids_linkoffset;
+
+ #ifdef HAVE_LIBGTHREAD_2_0
+     if(nids_params.multiproc) { 
+        /* 
+         * Insert received fragment into the async capture queue.
+         * We hope that the overhead of memcpy 
+         * will be saturated by the benefits of SMP - mcree
+         */
+        qitem=malloc(sizeof(struct cap_queue_item));
+        if (qitem && (qitem->data=malloc(hdr->caplen - nids_linkoffset))) {
+          qitem->caplen=hdr->caplen - nids_linkoffset;
+          memcpy(qitem->data,data_aligned,qitem->caplen);
+          g_async_queue_lock(cap_queue);
+          /* ensure queue does not overflow */
+          if(g_async_queue_length_unlocked(cap_queue) > nids_params.queue_limit) {
+	    /* queue limit reached: drop packet - should we notify user via syslog? */
+	    free(qitem->data);
+	    free(qitem);
+	    } else {
+	    /* insert packet to queue */
+	    g_async_queue_push_unlocked(cap_queue,qitem);
+          }
+          g_async_queue_unlock(cap_queue);
+	}
+     } else { /* user requested simple passthru - no threading */
+        call_ip_frag_procs(data_aligned,hdr->caplen - nids_linkoffset);
+     }
+ #else
+     call_ip_frag_procs(data_aligned,hdr->caplen - nids_linkoffset);
+ #endif
 }
 
 static void gen_ip_frag_proc(u_char * data, int len)
@@ -272,6 +350,8 @@ static void gen_ip_frag_proc(u_char * data, int len)
     struct ip *iph = (struct ip *) data;
     int need_free = 0;
     int skblen;
+    void (*glibc_syslog_h_workaround)(int, int, struct ip *, void*)=
+        nids_params.syslog;
 
     if (!nids_params.ip_filter(iph, len))
 	return;
@@ -279,11 +359,11 @@ static void gen_ip_frag_proc(u_char * data, int len)
     if (len < (int)sizeof(struct ip) || iph->ip_hl < 5 || iph->ip_v != 4 ||
 	ip_fast_csum((unsigned char *) iph, iph->ip_hl) != 0 ||
 	len < ntohs(iph->ip_len) || ntohs(iph->ip_len) < iph->ip_hl << 2) {
-	nids_params.syslog(NIDS_WARN_IP, NIDS_WARN_IP_HDR, iph, 0);
+	glibc_syslog_h_workaround(NIDS_WARN_IP, NIDS_WARN_IP_HDR, iph, 0);
 	return;
     }
     if (iph->ip_hl > 5 && ip_options_compile(data)) {
-	nids_params.syslog(NIDS_WARN_IP, NIDS_WARN_IP_SRR, iph, 0);
+	glibc_syslog_h_workaround(NIDS_WARN_IP, NIDS_WARN_IP_SRR, iph, 0);
 	return;
     }
     switch (ip_defrag_stub((struct ip *) data, &iph)) {
@@ -335,7 +415,8 @@ static void process_udp(char *data)
     ulen = ntohs(udph->UH_ULEN);
     if (len - hlen < ulen || ulen < (int)sizeof(struct udphdr))
 	return;
-    if (my_udp_check
+    /* According to RFC768 a checksum of 0 is not an error (Sebastien Raveau) */
+    if (udph->uh_sum && my_udp_check
 	((void *) udph, ulen, iph->ip_src.s_addr,
 	 iph->ip_dst.s_addr)) return;
     addr.source = ntohs(udph->UH_SPORT);
@@ -379,30 +460,34 @@ static void init_procs()
 
 void nids_register_udp(void (*x))
 {
-    struct proc_node *ipp = mknew(struct proc_node);
+    register_callback(&udp_procs, x);
+}
 
-    ipp->item = x;
-    ipp->next = udp_procs;
-    udp_procs = ipp;
+void nids_unregister_udp(void (*x))
+{
+    unregister_callback(&udp_procs, x);
 }
 
 void nids_register_ip(void (*x))
 {
-    struct proc_node *ipp = mknew(struct proc_node);
+    register_callback(&ip_procs, x);
+}
 
-    ipp->item = x;
-    ipp->next = ip_procs;
-    ip_procs = ipp;
+void nids_unregister_ip(void (*x))
+{
+    unregister_callback(&ip_procs, x);
 }
 
 void nids_register_ip_frag(void (*x))
 {
-    struct proc_node *ipp = mknew(struct proc_node);
-
-    ipp->item = x;
-    ipp->next = ip_frag_procs;
-    ip_frag_procs = ipp;
+    register_callback(&ip_frag_procs, x);
 }
+
+void nids_unregister_ip_frag(void (*x))
+{
+    unregister_callback(&ip_frag_procs, x);
+}
+
 static int open_live()
 {
     char *device;
@@ -438,9 +523,56 @@ static int open_live()
     return 1;
 }
 
+#ifdef HAVE_LIBGTHREAD_2_0
+
+#define START_CAP_QUEUE_PROCESS_THREAD() \
+    if(nids_params.multiproc) { /* threading... */ \
+	 if(!(g_thread_create_full((GThreadFunc)cap_queue_process_thread,NULL,0,FALSE,TRUE,G_THREAD_PRIORITY_LOW,&gerror))) { \
+	    strcpy(nids_errbuf, "thread: "); \
+	    strncat(nids_errbuf, gerror->message, sizeof(nids_errbuf) - 8); \
+	    return 0; \
+	 }; \
+    }
+
+#define STOP_CAP_QUEUE_PROCESS_THREAD() \
+    if(nids_params.multiproc) { /* stop the capture process thread */ \
+	 g_async_queue_push(cap_queue,&EOF_item); \
+    }
+
+
+/* thread entry point 
+ * pops capture queue items and feeds them to
+ * the ip fragment processors - mcree
+ */
+static void cap_queue_process_thread()
+{
+     struct cap_queue_item *qitem;
+     
+     while(1) { /* loop "forever" */
+	  qitem=g_async_queue_pop(cap_queue);
+	  if (qitem==&EOF_item) break; /* EOF item received: we should exit */
+	  call_ip_frag_procs(qitem->data,qitem->caplen);
+	  free(qitem->data);
+	  free(qitem);
+     }
+     g_thread_exit(NULL);
+}
+
+#else
+
+#define START_CAP_QUEUE_PROCESS_THREAD()
+#define STOP_CAP_QUEUE_PROCESS_THREAD()
+
+#endif
+
 int nids_init()
 {
-    if (nids_params.filename) {
+    /* free resources that previous usages might have allocated */
+    nids_exit();
+
+    if (nids_params.pcap_desc)
+        desc = nids_params.pcap_desc;
+    else if (nids_params.filename) {
 	if ((desc = pcap_open_offline(nids_params.filename,
 				      nids_errbuf)) == NULL)
 	    return 0;
@@ -467,36 +599,36 @@ int nids_init()
 #endif
 #ifdef DLT_NULL
     case DLT_NULL:
-        linkoffset = 4;
+        nids_linkoffset = 4;
         break;
 #endif        
     case DLT_EN10MB:
-	linkoffset = 14;
+	nids_linkoffset = 14;
 	break;
     case DLT_PPP:
-	linkoffset = 4;
+	nids_linkoffset = 4;
 	break;
 	/* Token Ring Support by vacuum@technotronic.com, thanks dugsong! */
     case DLT_IEEE802:
-	linkoffset = 22;
+	nids_linkoffset = 22;
 	break;
 
     case DLT_RAW:
     case DLT_SLIP:
-	linkoffset = 0;
+	nids_linkoffset = 0;
 	break;
 #define DLT_LINUX_SLL   113
     case DLT_LINUX_SLL:
-	linkoffset = 16;
+	nids_linkoffset = 16;
 	break;
 #ifdef DLT_FDDI
     case DLT_FDDI:
-        linkoffset = 21;
+        nids_linkoffset = 21;
         break;
 #endif        
 #ifdef DLT_PPP_SERIAL 
     case DLT_PPP_SERIAL:
-        linkoffset = 4;
+        nids_linkoffset = 4;
         break;
 #endif        
     default:
@@ -516,20 +648,57 @@ int nids_init()
     tcp_init(nids_params.n_tcp_streams);
     ip_frag_init(nids_params.n_hosts);
     scan_init();
+
+    if(nids_params.multiproc) {
+#ifdef HAVE_LIBGTHREAD_2_0
+	 g_thread_init(NULL);
+	 cap_queue=g_async_queue_new();
+#else
+	 strcpy(nids_errbuf, "libnids was compiled without threads support");
+	 return 0;        
+#endif
+    }
+
     return 1;
 }
 
-void nids_run()
+int nids_run()
 {
     if (!desc) {
 	strcpy(nids_errbuf, "Libnids not initialized");
+	return 0;
+    }
+    START_CAP_QUEUE_PROCESS_THREAD(); /* threading... */
+    pcap_loop(desc, -1, (pcap_handler) nids_pcap_handler, 0);
+    /* FIXME: will this code ever be called? Don't think so - mcree */
+    STOP_CAP_QUEUE_PROCESS_THREAD(); 
+    nids_exit();
+    return 0;
+}
+
+void nids_exit()
+{
+    if (!desc) {
+        strcpy(nids_errbuf, "Libnids not initialized");
 	return;
     }
-    pcap_loop(desc, -1, (pcap_handler) pcap_hand, 0);
-    clear_stream_buffers();
+#ifdef HAVE_LIBGTHREAD_2_0
+    if (nids_params.multiproc) {
+    /* I have no portable sys_sched_yield,
+       and I don't want to add more synchronization...
+    */
+      while (g_async_queue_length(cap_queue)>0) 
+        usleep(100000);
+    }
+#endif
+    tcp_exit();
+    ip_frag_exit();
+    scan_exit();
     strcpy(nids_errbuf, "loop: ");
-    strncat(nids_errbuf, pcap_geterr(desc), sizeof(nids_errbuf) - 7);
-    pcap_close(desc);
+    strncat(nids_errbuf, pcap_geterr(desc), sizeof nids_errbuf - 7);
+    if (!nids_params.pcap_desc)
+        pcap_close(desc);
+    desc = NULL;
 }
 
 int nids_getfd()
@@ -555,7 +724,10 @@ int nids_next()
 	strncat(nids_errbuf, pcap_geterr(desc), sizeof(nids_errbuf) - 7);
 	return 0;
     }
-    pcap_hand(0, &h, data);
+    /* threading is quite useless (harmful) in this case - should we do an API change?  */
+    START_CAP_QUEUE_PROCESS_THREAD();
+    nids_pcap_handler(0, &h, data);
+    STOP_CAP_QUEUE_PROCESS_THREAD();
     return 1;
 }
 
@@ -567,10 +739,12 @@ int nids_dispatch(int cnt)
 	strcpy(nids_errbuf, "Libnids not initialized");
 	return -1;
     }
-    if ((r = pcap_dispatch(desc, cnt, (pcap_handler) pcap_hand,
+    START_CAP_QUEUE_PROCESS_THREAD(); /* threading... */
+    if ((r = pcap_dispatch(desc, cnt, (pcap_handler) nids_pcap_handler,
                                     NULL)) == -1) {
 	strcpy(nids_errbuf, "dispatch: ");
 	strncat(nids_errbuf, pcap_geterr(desc), sizeof(nids_errbuf) - 11);
     }
+    STOP_CAP_QUEUE_PROCESS_THREAD(); 
     return r;
 }
