@@ -57,6 +57,7 @@ static int max_stream;
 static struct tcp_stream *tcp_latest = 0, *tcp_oldest = 0;
 static struct tcp_stream *free_streams;
 static struct ip *ugly_iphdr;
+struct tcp_timeout *nids_tcp_timeouts = 0;
 
 static void purge_queue(struct half_stream * h)
 {
@@ -73,11 +74,62 @@ static void purge_queue(struct half_stream * h)
 }
 
 static void
-free_tcp(struct tcp_stream * a_tcp)
+add_tcp_closing_timeout(struct tcp_stream * a_tcp)
+{
+  struct tcp_timeout *to;
+  struct tcp_timeout *newto;
+
+  if (!nids_params.tcp_workarounds)
+    return;
+  newto = malloc(sizeof (struct tcp_timeout));
+  newto->a_tcp = a_tcp;
+  newto->timeout.tv_sec = nids_last_pcap_header->ts.tv_sec + 10;
+  newto->prev = 0;
+  for (newto->next = to = nids_tcp_timeouts; to; newto->next = to = to->next) {
+    if (to->a_tcp == a_tcp) {
+      free(newto);
+      return;
+    }
+    if (to->timeout.tv_sec > newto->timeout.tv_sec)
+      break;
+    newto->prev = to;
+  }
+  if (!newto->prev)
+    nids_tcp_timeouts = newto;
+  else
+    newto->prev->next = newto;
+  if (newto->next)
+    newto->next->prev = newto;
+}
+
+static void
+del_tcp_closing_timeout(struct tcp_stream * a_tcp)
+{
+  struct tcp_timeout *to;
+
+  if (!nids_params.tcp_workarounds)
+    return;
+  for (to = nids_tcp_timeouts; to; to = to->next)
+    if (to->a_tcp == a_tcp)
+      break;
+  if (!to)
+    return;
+  if (!to->prev)
+    nids_tcp_timeouts = to->next;
+  else
+    to->prev->next = to->next;
+  if (to->next)
+    to->next->prev = to->prev;
+  free(to);
+}
+
+void
+nids_free_tcp_stream(struct tcp_stream * a_tcp)
 {
   int hash_index = a_tcp->hash_index;
   struct lurker_node *i, *j;
- 
+
+  del_tcp_closing_timeout(a_tcp);
   purge_queue(&a_tcp->server);
   purge_queue(&a_tcp->client);
    
@@ -112,6 +164,24 @@ free_tcp(struct tcp_stream * a_tcp)
   tcp_num--;
 }
 
+void
+tcp_check_timeouts(struct timeval *now)
+{
+  struct tcp_timeout *to;
+  struct tcp_timeout *next;
+  struct lurker_node *i;
+
+  for (to = nids_tcp_timeouts; to; to = next) {
+    if (now->tv_sec < to->timeout.tv_sec)
+      return;
+    to->a_tcp->nids_state = NIDS_TIMED_OUT;
+    for (i = to->a_tcp->listeners; i; i = i->next)
+      (i->item) (to->a_tcp, &i->data);
+    next = to->next;
+    nids_free_tcp_stream(to->a_tcp);
+  }
+}
+
 static int
 mk_hash_index(struct tuple4 addr)
 {
@@ -195,11 +265,11 @@ add_new_tcp(struct tcphdr * this_tcphdr, struct ip * this_iphdr)
   
   if (tcp_num > max_stream) {
     struct lurker_node *i;
-    
+
     tcp_oldest->nids_state = NIDS_TIMED_OUT;
     for (i = tcp_oldest->listeners; i; i = i->next)
       (i->item) (tcp_oldest, &i->data);
-    free_tcp(tcp_oldest);
+    nids_free_tcp_stream(tcp_oldest);
     nids_params.syslog(NIDS_WARN_TCP, NIDS_WARN_TCP_TOOMUCH, ugly_iphdr, this_tcphdr);
   }
   a_tcp = free_streams;
@@ -277,7 +347,7 @@ ride_lurkers(struct tcp_stream * a_tcp, char mask)
       sc = a_tcp->server.collect;
       ccu = a_tcp->client.collect_urg;
       scu = a_tcp->server.collect_urg;
-      
+
       (i->item) (a_tcp, &i->data);
       if (cc < a_tcp->client.collect)
 	i->whatto |= COLLECT_cc;
@@ -407,8 +477,11 @@ add_from_skb(struct tcp_stream * a_tcp, struct half_stream * rcv,
       }
     }
   }
-  if (fin)
+  if (fin) {
     snd->state = FIN_SENT;
+    if (rcv->state == TCP_CLOSING)
+      add_tcp_closing_timeout(a_tcp);
+  }
 }
 
 static void
@@ -479,6 +552,19 @@ tcp_queue(struct tcp_stream * a_tcp, struct tcphdr * this_tcphdr,
       nids_params.no_mem("tcp_queue");
     memcpy(pakiet->data, data, datalen);
     pakiet->fin = (this_tcphdr->th_flags & TH_FIN);
+    /* Some Cisco - at least - hardware accept to close a TCP connection
+     * even though packets were lost before the first TCP FIN packet and
+     * never retransmitted; this violates RFC 793, but since it really
+     * happens, it has to be dealt with... The idea is to introduce a 10s
+     * timeout after TCP FIN packets were sent by both sides so that
+     * corresponding libnids resources can be released instead of waiting
+     * for retransmissions which will never happen.  -- Sebastien Raveau
+     */
+    if (pakiet->fin) {
+      snd->state = TCP_CLOSING;
+      if (rcv->state == FIN_SENT || rcv->state == FIN_CONFIRMED)
+	add_tcp_closing_timeout(a_tcp);
+    }
     pakiet->seq = this_seq;
     pakiet->urg = (this_tcphdr->th_flags & TH_URG);
     pakiet->urg_ptr = ntohs(this_tcphdr->th_urp);
@@ -550,17 +636,13 @@ find_stream(struct tcphdr * this_tcphdr, struct ip * this_iphdr,
 	    int *from_client)
 {
   struct tuple4 this_addr, reversed;
-  int hash_index;
   struct tcp_stream *a_tcp;
 
   this_addr.source = ntohs(this_tcphdr->th_sport);
   this_addr.dest = ntohs(this_tcphdr->th_dport);
   this_addr.saddr = this_iphdr->ip_src.s_addr;
   this_addr.daddr = this_iphdr->ip_dst.s_addr;
-  hash_index = mk_hash_index(this_addr);
-  for (a_tcp = tcp_stream_table[hash_index];
-       a_tcp && !b_comp(a_tcp->addr, this_addr);
-       a_tcp = a_tcp->next_node);
+  a_tcp = nids_find_tcp_stream(&this_addr);
   if (a_tcp) {
     *from_client = 1;
     return a_tcp;
@@ -569,36 +651,53 @@ find_stream(struct tcphdr * this_tcphdr, struct ip * this_iphdr,
   reversed.dest = ntohs(this_tcphdr->th_sport);
   reversed.saddr = this_iphdr->ip_dst.s_addr;
   reversed.daddr = this_iphdr->ip_src.s_addr;
-  hash_index = mk_hash_index(reversed);
-  for (a_tcp = tcp_stream_table[hash_index];
-       a_tcp && !b_comp(a_tcp->addr, reversed);
-       a_tcp = a_tcp->next_node);
+  a_tcp = nids_find_tcp_stream(&reversed);
   if (a_tcp) {
     *from_client = 0;
     return a_tcp;
   }
-  else
-    return 0;
+  return 0;
 }
 
-void clear_stream_buffers()
+struct tcp_stream *
+nids_find_tcp_stream(struct tuple4 *addr)
+{
+  int hash_index;
+  struct tcp_stream *a_tcp;
+
+  hash_index = mk_hash_index(*addr);
+  for (a_tcp = tcp_stream_table[hash_index];
+       a_tcp && memcmp(&a_tcp->addr, addr, sizeof (struct tuple4));
+       a_tcp = a_tcp->next_node);
+  return a_tcp ? a_tcp : 0;
+}
+
+
+void tcp_exit(void)
 {
   int i;
   struct lurker_node *j;
   struct tcp_stream *a_tcp;
-  
+
+  if (!tcp_stream_table || !streams_pool)
+    return;
   for (i = 0; i < tcp_stream_table_size; i++) {
     for (a_tcp = tcp_stream_table[i];
          a_tcp;
          a_tcp = a_tcp->next_node) {
       for (j = a_tcp->listeners; j; j = j->next) {
           a_tcp->nids_state = NIDS_EXITING;
-          (j->item)(a_tcp, &j->data);
+	  (j->item)(a_tcp, &j->data);
       }
     }
   }
+  free(tcp_stream_table);
+  tcp_stream_table = NULL;
+  free(streams_pool);
+  streams_pool = NULL;
+  /* FIXME: anything else we should free? */
 }
-        
+
 void
 process_tcp(u_char * data, int skblen)
 {
@@ -700,12 +799,12 @@ process_tcp(u_char * data, int skblen)
   if ((this_tcphdr->th_flags & TH_RST)) {
     if (a_tcp->nids_state == NIDS_DATA) {
       struct lurker_node *i;
-      
+
       a_tcp->nids_state = NIDS_RESET;
       for (i = a_tcp->listeners; i; i = i->next)
 	(i->item) (a_tcp, &i->data);
     }
-    free_tcp(a_tcp);
+    nids_free_tcp_stream(a_tcp);
     return;
   }
 
@@ -763,7 +862,7 @@ process_tcp(u_char * data, int skblen)
 	    }
 	  }
 	  if (!a_tcp->listeners) {
-	    free_tcp(a_tcp);
+	    nids_free_tcp_stream(a_tcp);
 	    return;
 	  }
 	  a_tcp->nids_state = NIDS_DATA;
@@ -778,11 +877,11 @@ process_tcp(u_char * data, int skblen)
       rcv->state = FIN_CONFIRMED;
     if (rcv->state == FIN_CONFIRMED && snd->state == FIN_CONFIRMED) {
       struct lurker_node *i;
-      
+
       a_tcp->nids_state = NIDS_CLOSE;
       for (i = a_tcp->listeners; i; i = i->next)
 	(i->item) (a_tcp, &i->data);
-      free_tcp(a_tcp);
+      nids_free_tcp_stream(a_tcp);
       return;
     }
   }
@@ -794,7 +893,7 @@ process_tcp(u_char * data, int skblen)
   if (rcv->rmem_alloc > 65535)
     prune_queue(rcv, this_tcphdr);
   if (!a_tcp->listeners)
-    free_tcp(a_tcp);
+    nids_free_tcp_stream(a_tcp);
 }
 
 void
@@ -807,33 +906,44 @@ nids_discard(struct tcp_stream * a_tcp, int num)
 void
 nids_register_tcp(void (*x))
 {
-  struct proc_node *i;
+  register_callback(&tcp_procs, x);
+}
 
-  i = mknew(struct proc_node);
-  i->item = x;
-  i->next = tcp_procs;
-  tcp_procs = i;
+void
+nids_unregister_tcp(void (*x))
+{
+  unregister_callback(&tcp_procs, x);
 }
 
 int
 tcp_init(int size)
 {
   int i;
+  struct tcp_timeout *tmp;
+
   if (!size) return 0;
   tcp_stream_table_size = size;
-  tcp_stream_table = malloc(tcp_stream_table_size * sizeof(char *));
-  if (!tcp_stream_table)
+  tcp_stream_table = calloc(tcp_stream_table_size, sizeof(char *));
+  if (!tcp_stream_table) {
     nids_params.no_mem("tcp_init");
-  memset(tcp_stream_table, 0, tcp_stream_table_size * sizeof(char *));
+    return -1;
+  }
   max_stream = 3 * tcp_stream_table_size / 4;
   streams_pool = (struct tcp_stream *) malloc((max_stream + 1) * sizeof(struct tcp_stream));
-  if (!streams_pool)
+  if (!streams_pool) {
     nids_params.no_mem("tcp_init");
+    return -1;
+  }
   for (i = 0; i < max_stream; i++)
     streams_pool[i].next_free = &(streams_pool[i + 1]);
   streams_pool[max_stream].next_free = 0;
   free_streams = streams_pool;
-  init_hash(); 
+  init_hash();
+  while (nids_tcp_timeouts) {
+    tmp = nids_tcp_timeouts->next;
+    free(nids_tcp_timeouts);
+    nids_tcp_timeouts = tmp;
+  }
   return 0;
 }
 
@@ -915,5 +1025,5 @@ process_icmp(u_char * data)
   a_tcp->nids_state = NIDS_RESET;
   for (i = a_tcp->listeners; i; i = i->next)
     (i->item) (a_tcp, &i->data);
-  free_tcp(a_tcp);
+  nids_free_tcp_stream(a_tcp);
 }
